﻿// Deathstrike CC
// DeathKnight Release
// by Wulf

// Credits to fiftypence from the Honorbuddy forums his DoctrineRet CC without which this CC would not be possible.


using System; //added
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Threading;
using Styx;
using Styx.Combat.CombatRoutine;
using Styx.Helpers;
using Styx.Logic;
using Styx.Logic.Combat;
using Styx.Logic.Pathing;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

namespace Deathstrike
{
    partial class DeathKnight : CombatRoutine
    {
        // The following are common functions, lists, etc.
        // They will be used throughout Deathstrike.


        public List<WoWUnit> EnemyUnits
        {
            get
            {
                return
                    ObjectManager.GetObjectsOfType<WoWUnit>(false, false)
                    .Where(unit =>
                        !unit.IsFriendly
                        && (unit.IsTargetingMeOrPet
                            || unit.IsTargetingMyPartyMember)
                        && unit.IsAlive
                        && !unit.IsPet
                        && !unit.IsNonCombatPet
                        && !unit.IsCritter
                        && unit.DistanceSqr
                     <= 10 * 10).ToList();
            }
        }

        public List<WoWUnit> EnemyUnitsTauntable
        {
            get
            {
                return
                    ObjectManager.GetObjectsOfType<WoWUnit>(false, false)
                    .Where(unit =>
                        !unit.IsFriendly
                        && unit.IsTargetingMyPartyMember
                        && unit.IsAlive
                        && unit.ThreatInfo.RawPercent < 100
                        && !unit.IsTargetingMeOrPet
                        && !unit.IsPet
                        && !unit.IsNonCombatPet
                        && !unit.IsCritter
                        && unit.DistanceSqr
                    <= 80 * 80).ToList();
            }
        }

        public List<WoWPlayer> FriendlyPlayers
        {
            get
            {
                return
                    ObjectManager.GetObjectsOfType<WoWPlayer>(false, false)
                    .Where(unit =>
                        unit.IsInMyPartyOrRaid
                        && !unit.IsPet
                        && unit.DistanceSqr
                    <= 40 * 40).ToList();
            }
        }

        public WoWUnit FindClosestTarget()
        {
            return (from unit in EnemyUnits
                    where unit.IsAlive
                    orderby unit.Distance2D ascending
                    select unit).FirstOrDefault();
        }

        public WoWUnit FindLowestThreat()
        {
            return (from unit in EnemyUnits
                    where unit.IsAlive
                    orderby unit.ThreatInfo.RawPercent ascending
                    select unit).FirstOrDefault();
        }

        public WoWUnit FindTauntTarget()
        {
            return (from unit in EnemyUnitsTauntable
                    where unit.IsAlive
                    orderby unit.Distance2D descending
                    select unit).FirstOrDefault();
        }

        public void GetTarget()
        {
            // choose new target if current target is null (checked in combat())
            if (UseTarget == 1)
            {
                WoWUnit LowestThreat = FindLowestThreat();

                if (LowestThreat != null)
                {                   
                    LowestThreat.Target();
                }
            }
            else if (UseTarget == 2)
            {
                WoWUnit ClosestTarget = FindClosestTarget();

                if (ClosestTarget != null)
                {
                    ClosestTarget.Target();
                }
            }
            else
            {
                // this probably only works with lazyraider
                if (RaFHelper.Leader != null && RaFHelper.Leader.CurrentTarget != null)
                {
                    RaFHelper.Leader.CurrentTarget.Target();
                }
            }
        }

        public bool IsCrowdControlled(WoWUnit unit)
        {
            Dictionary<string, WoWAura>.ValueCollection auras = unit.Auras.Values;

            return auras.Any(
                a => a.Spell.Mechanic == WoWSpellMechanic.Banished ||
                     a.Spell.Mechanic == WoWSpellMechanic.Charmed ||
                     a.Spell.Mechanic == WoWSpellMechanic.Horrified ||
                     a.Spell.Mechanic == WoWSpellMechanic.Incapacitated ||
                     a.Spell.Mechanic == WoWSpellMechanic.Polymorphed ||
                     a.Spell.Mechanic == WoWSpellMechanic.Sapped ||
                     a.Spell.Mechanic == WoWSpellMechanic.Shackled ||
                     a.Spell.Mechanic == WoWSpellMechanic.Asleep ||
                     a.Spell.Mechanic == WoWSpellMechanic.Frozen
                     );
        }
        
        public bool damageReduction()
        {

            if (Presence == 1) //Blood Presence form single target
            {
                if (Me.HealthPercent < VampiricBlood)
                {
                    if (SpellManager.CanCast("Vampiric Blood") && !SpellManager.CanCast("Death Strike"))
                    {
                        if (logDebug)
                        {
                            logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Vampiric Blood", 0, 0, false, BuffExists("Will of the Necropolis"));
                        }

                        SpellManager.Cast("Vampiric Blood");
                        return false;
                    }

                }
                if (Me.HealthPercent < RuneTap)
                {
                    if (SpellManager.CanCast("Rune Tap") && BuffExists("Will of the Necropolis") && !SpellManager.CanCast("Death Strike") && !IsOnCooldown("Rune Tap"))
                    {
                        if (logDebug)
                        {
                            logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Rune Tap", 0, 0, false, BuffExists("Will of the Necropolis"));
                        }

                        SpellManager.Cast("Rune Tap");
                        return false;
                    }
                    else if ((Me.BloodRuneCount > 0 || Me.DeathRuneCount > 0) && !SpellManager.CanCast("Death Strike") && !IsOnCooldown("Rune Tap"))
                    {
                        if (logDebug)
                        {
                            logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Rune Tap", 0, 0, false, BuffExists("Will of the Necropolis"));
                        }

                        SpellManager.Cast("Rune Tap");
                        return false;
                    }

                }
                if (!SpellManager.GlobalCooldown)
                {
                    // Lichborne - Instant / 2 minute cooldown / requires runic power for death coils.
                    if (Me.CurrentRunicPower > 40 && Me.HealthPercent < LichborneDC)
                    {
                        if ((SpellManager.CanCast("Lichborne")) && (SpellManager.CanCast("Death Coil")) && !SpellManager.CanCast("Death Strike"))
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Lichborne", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }

                            SpellManager.Cast("Lichborne");
                            SpellManager.Cast("Deathcoil", Me);
                            return false;

                        }
                        else if ((!SpellManager.CanCast("Lichborne")) && (SpellManager.CanCast("Death Coil")) && !SpellManager.CanCast("Death Strike") && BuffExists("Lichborne"))
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Lichborne & Deathcoil", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }

                            SpellManager.Cast("Deathcoil", Me);
                            return false;

                        }
                    }

                    if (!Me.HasAura("Bone Shield"))
                    {
                        if (SpellManager.CanCast("Blood Tap") && !IsOnCooldown("Bone Shield"))
                        {
                            SpellManager.Cast("Blood Tap");
                            Logging.Write(Color.Lime, "=== [" + Me.CurrentHealth + " Hp] Blood Tap for Bone Shield");
                                                       
                            SpellManager.Cast("Bone Shield");
                            Logging.Write(Color.Lime, "=== [" + Me.CurrentHealth + " Hp] Bone Shield Activated");
                            return false;
                        }
                        if ((Me.UnholyRuneCount > 0 || Me.DeathRuneCount > 0) && !IsOnCooldown("Bone Shield"))
                        {
                            SpellManager.Cast("Bone Shield");
                            Logging.Write(Color.Lime, "=== [" + Me.CurrentHealth + " Hp] Bone Shield Activated");
                            return false;

                        }

                    }
                }
            
            }

            /////////////////////////////////////////////////////////////////////////////////////////////////////////////
            /////     End Blood Specific Life Savers
            /////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            // Icebound Fortitude - Instant / 3 minute cooldown / Reduces all damage by 50% for 12 seconds.
            if (Me.CurrentRunicPower >= 20 && Me.HealthPercent < IceboundFortitude)
            {
                if (SpellManager.CanCast("Icebound Fortitude") && !SpellManager.CanCast("Death Strike"))
                {
                    if (logDebug)
                    {
                        logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Icebound Fortitude", 0, 0, false, BuffExists("Will of the Necropolis"));
                    }

                    SpellManager.Cast("Icebound Fortitude");
                    return false;
                }

            }
			 // Anti-Magic Shell
             if ((Me.CurrentTarget.IsCasting || Me.CurrentTarget.ChanneledCastingSpellId != 0) && (SpellManager.CanCast("Anti-Magic Shell")))
             {
                    Logging.Write(Color.Lime, "=== [" + Me.CurrentHealth + " Hp] Anti-Magic Shell");
                    
                    SpellManager.Cast("Anti-Magic Shell");
                    return false;
            }
            if (!SpellManager.GlobalCooldown)
            {
                /////////////////////////////////////////////////////////////////////////////////////////////////////////////
                /////     Life Savers
                /////////////////////////////////////////////////////////////////////////////////////////////////////////////

                //Death Pact - 40 Runic power / 2 minute cooldown / Heals 25% maximum health
                if (Me.CurrentRunicPower > 40 && Me.HealthPercent < DeathPact)
                {
                    if ((SpellManager.CanCast("Death Pact")) && (SpellManager.CanCast("Raise Dead")) && !SpellManager.CanCast("Death Strike"))
                    {
                        if (logDebug)
                        {
                            logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Death Pact", 0, 0, false, BuffExists("Will of the Necropolis"));
                        }

                        SpellManager.Cast("Raise Dead");
                        Thread.Sleep(2000);
                        SpellManager.Cast("Death Pact");
                        return false;

                    }
                }

                // Keep Horn of winter up.
                if (!Me.Auras.ContainsKey("Horn of Winter") && !Me.Auras.ContainsKey("Battle Shout") && !Me.Auras.ContainsKey("Roar of Courage") && !Me.Auras.ContainsKey("Strength of Earth Totem") && SpellManager.CanCast("Horn of Winter") && PlayerBuffTimeLeft("Horn of Winter") <= 2)
                {
                    Logging.Write(Color.Lime, "=== [" + Me.CurrentHealth + " Hp] Horn of Winter");
                    
                    SpellManager.Cast("Horn of Winter");
                    return false;
                }

				
                if (ForcePresence)
                {
                    if (Presence == 0)
                    {
                        if (!Me.HasAura("Unholy Presence")
                            && SpellManager.CanCast("Unholy Presence"))
                        {
                            Logging.Write(Color.Lime, "=== [" + Me.Name + "] Switching to Unholy Presence");

                            SpellManager.Cast("Unholy Presence");
                            return false;
                        }
                    }
                    else if (Presence == 1)
                    {
                        if (!Me.HasAura("Blood Presence")
                            && SpellManager.CanCast("Blood Presence"))
                        {
                            Logging.Write(Color.Lime, "=== [" + Me.Name + "] Switching to Blood Presence");
                            
                            SpellManager.Cast("Blood Presence");
                            return false;
                        }
                    }
                }


                // Anti-Magic Spell - Instant / 45 second cooldown / absorbs 75% magic damage for 5 secs.
            }
            return false; // end damage reduction        
        }

        // May need to follow this up with an Icy Touch to get the adds back...
        public void Taunt()
        {
            WoWUnit TauntTarget = FindTauntTarget();
            /*if (Me.Combat)
            {
                if (Me.IsInInstance)
                {
                    foreach (WoWUnit unit in ObjectManager.GetObjectsOfType<WoWUnit>())
                    {
                        Logging.Write("=====================[Threat Info]============================");
                        if (!unit.Combat) continue;
                        if (unit.Distance > 80) continue;
                        Logging.Write(string.Format("-- {0}, TargetGuid {1}, ThreatStatus {2}, ThreatPercent {3}, RawPercent {4}", unit.Name, unit.ThreatInfo.TargetGuid, unit.ThreatInfo.ThreatStatus, unit.ThreatInfo.RawPercent, unit.ThreatInfo.RawPercent));
                        Logging.Write("-- Threat for ME " + unit.GetThreatInfoFor(Me));

                    }                   
                }
            }*/
            if (TauntTarget != null && UseTaunt == true)
            {
                StyxWoW.SleepForLagDuration();

                if (TauntTarget.Distance2D < MaxDistance("Dark Command")
                    && SpellManager.CanCast("Dark Command") && !TauntTarget.HasAura("Dark Command"))
                {
                    Logging.Write(Color.Red, "=== [" + Me.CurrentTarget.ThreatInfo.RawPercent + "%] Dark Command on " + TauntTarget.Name);
                    
                    SpellManager.Cast("Dark Command", TauntTarget);
                    return;
                }
                //Death Grip's taunt cannot miss, but mobs can be immune to the grip part of the spell. 
                else if (TauntTarget.Distance2D < MaxDistance("Death Grip")
                    && SpellManager.CanCast("Death Grip") && !TauntTarget.HasAura("Dark Command"))
                {
                    Logging.Write(Color.Red, "=== [" + Me.CurrentTarget.ThreatInfo.RawPercent + "%] Death Grip on " + TauntTarget.Name);
                    
                    SpellManager.Cast("Death Grip", TauntTarget);
                    return;
                }
            }
        }

        //Death Knights have some passive threat reduction 24%?
        public void ReduceThreat()
        {
            if (Me.CurrentTarget.ThreatInfo.RawPercent >= 95 && Me.IsInInstance)
            {                           
                WoWUnit LowestThreat = FindLowestThreat();
                Logging.Write(Color.Red, "=== [" + Me.CurrentTarget.ThreatInfo.RawPercent + "%] Threat on " + Me.CurrentTarget.Name);

                if (LowestThreat != null && EnemyUnits.Count > 1)
                {
                    LowestThreat.Target();
                    Logging.Write(Color.Red, "=== [" + Me.CurrentTarget.ThreatInfo.RawPercent + "%] " + Me.CurrentTarget.Name + " is new Target");
                    return;
                }                               
            }
        }

        //TODO: Check they are speced into Endless winter. This reduces the Runic power cost.
        public void Interrupt()
        {
            if (UseInterrupt)
            {
            RunLUA("local u,i=UnitCastingInfo;i=select(9,u\"target\")if u\"target\"and not i and IsSpellInRange(\"Mind Freeze\",\"target\")==1 then CastSpellByName(\"Mind Freeze\",\"target\")end");
            }
        }

        private void AutoAttack()
        {
            if (!Me.IsAutoAttacking && !Me.Mounted)
            {
                Lua.DoString("StartAttack()");
            }

        }        
        
        public void SingleTargetRotation()
        {
            if (Presence == 1) //Blood Presence single target
            {
                // --- single target PvE dps rotation

                /* Theory crafting here.
                  
                  actions+=/presence,choose=Blood
                  actions+=/army_of_the_dead
                  actions+=/golemblood_potion,if=!in_combat|buff.bloodlust.react|target.time_to_die<=60
                  actions+=/auto_attack
                  actions+=/bone_shield,if=!buff.bone_shield.up
                  actions+=/icy_touch,if=dot.frost_fever.remains<=2
                  actions+=/plague_strike,if=dot.blood_plague.remains<=2
                  actions+=/heart_strike
                  actions+=/death_strike
                  actions+=/dancing_rune_weapon,time<=150,if=dot.frost_fever.remains<=5|dot.blood_plague.remains<=5
                  actions+=/dancing_rune_weapon,if=(dot.frost_fever.remains<=5|dot.blood_plague.remains<=5)&buff.bloodlust.react
                  actions+=/empower_rune_weapon,if=blood=0&unholy=0&frost=0
                  actions+=/death_coil 
                
                  1) Deathstrike if damage taken within last 5 secs > A and Health < B %
                  2) Apply/refresh Outbreak/Plagustrike/Icy touch as debuffs are "needed"
                  3) Deathstrike if all frost and unholy runes or all death runes are full and you are below C% health or you need to refresh a shield with more than D% health or you have taken E% damage within the last 5 seconds
                  4) Heartstrike/Blood Boil if blood Runes are full or bladebarrier is down or bladebarrier expires in 3 seconds or the 2nd blood rune is about to refresh.
                  5) Bloodtap unless it has to be saved for a later Boneshield, Rune Tap, DnD
                  6) Runestrike
                  7) DRW
                  8) Horn of Winter
                  9) Deathcoil
                

                  */

                // Instant Spells here, this includes any procs that allow a free ability to be used.

                AutoAttack();               

                // Global cooldown restricted abilitiy's here
                if (SpellManager.GlobalCooldownLeft.Seconds < 0.5)
                {
                    //Death Strike  - 1 Unholy & 1 Frost / melee range
                    if (((Me.FrostRuneCount > 0 && Me.UnholyRuneCount > 0) || Me.DeathRuneCount >= 2) && IsWithinInteractRange)
                    {                                                   
                        if(logDebug)
                        {
                            logWrite(Color.Orange, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Death Strike", 0, 0, false, BuffExists("Will of the Necropolis"));                       
                        }

                        if (Me.HealthPercent < 80)
                        {
                            SpellManager.Cast("Death Strike");                            
                            return;
                        }
                        else if (Me.HealthPercent > 80 && BuffExists("Blood Shield") && PlayerBuffTimeLeft("Blood Shield") < 5)
                        {
                            SpellManager.Cast("Death Strike");
                            return;
                        }
                        else
                        {
                            SpellManager.Cast("Death Strike");
                            return;
                        }
                    }
                    else 
                    {

                        //Outbreak to apply Frost Fever and Blood Plague and Scarlet Fever - instant / 30 yd range 1 minute cooldown
                        if (!IsOnCooldown("Outbreak") && (!IsDebuffOnTarget("Frost Fever") && !IsDebuffOnTarget("Blood Plague")) && IsDistanceLessThan(MaxDistance("Outbreak")) && Me.CurrentTarget.HealthPercent > 20)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Orange, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Outbreak", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }
                            
                            SpellManager.Cast("Outbreak");
                            return;
                        }

                        //Icy Touch to apply Frost Fever - 1 Frost / 20 yd range
                        if ((Me.FrostRuneCount > 0 || Me.DeathRuneCount > 0)  && !IsDebuffOnTarget("Frost Fever") && IsDistanceLessThan(20) && Me.CurrentTarget.HealthPercent > 20 && Me.HealthPercent > 30)
                        {

                            if (TargetDebuffTimeLeft("Frost Fever") <= 4 || !IsOnCooldown("Icy Touch"))
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.Orange, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Icy Touch", 0, 0, false, BuffExists("Will of the Necropolis"));
                                }
                                
                                SpellManager.Cast("Icy Touch");
                                return;
                            }

                        }

                        //Plague Strike to apply Blood Plague and Scarlet Fever - 1 Unholy / melee range
                        if ((Me.UnholyRuneCount > 0 || Me.DeathRuneCount > 0) && !IsDebuffOnTarget("Blood Plague") && IsWithinInteractRange && Me.CurrentTarget.HealthPercent > 20 && Me.HealthPercent > 30)
                        {
                            if (TargetDebuffTimeLeft("Blood Plague") <= 4 || !IsOnCooldown("Plague Strike"))
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.Orange, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Plague Strike", 0, 0, false, BuffExists("Will of the Necropolis"));
                                }

                                SpellManager.Cast("Plague Strike");
                                return;
                            }
                        }

                        //Heart Strike  - 1 Blood / melee range
                        if ((Me.BloodRuneCount > 1 || Me.DeathRuneCount > 0) && !IsOnCooldown("Heart Strike") && IsWithinInteractRange)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Orange, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Heart Strike", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }
                            
                            SpellManager.Cast("Heart Strike");                         
                            return;
                        }


                        //Rune Strike  - 30 Runic Power / melee range
                        if (Me.RunicPowerPercent > 25 && IsWithinInteractRange && Me.HealthPercent > 30)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Orange, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Rune Strike", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }
                            
                            SpellManager.Cast("Rune Strike");
                            return;
                        }

                        //Death Coil  - 40 Runic Power / 30 yards
                        if (Me.RunicPowerPercent > 35 && IsDistanceLessThan(MaxDistance("Death Coil")) && !BuffExists("Lichborne") && Me.HealthPercent > 30)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Orange, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Death Coil", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }

                            SpellManager.Cast("Death Coil");
                            return;
                        }
                    }
                
                } // end global spell cooldown
            } // End Blood Presence

            if (Presence == 0)//Unholy Presence - Frost spec uses Unholy presence single target
            {
                /*
                glyphs=obliterate/frost_strike/howling_blast/pestilence/blood_boil/dark_succor/horn_of_winter/blood_tap/path_of_frost
                actions=flask,type=titanic_strength
                actions+=/food,type=beer_basted_crocolisk
                actions+=/presence,choose=unholy
                actions+=/army_of_the_dead
                actions+=/snapshot_stats
                actions+=/golemblood_potion,if=!in_combat|buff.bloodlust.react|target.time_to_die<=60
                actions+=/auto_attack
                actions+=/pillar_of_frost
                actions+=/blood_tap,if=death!=2
                actions+=/raise_dead,if=buff.rune_of_the_fallen_crusader.react&buff.heart_of_rage.react
                actions+=/raise_dead,time>=15
                actions+=/outbreak,if=dot.frost_fever.remains<=2|dot.blood_plague.remains<=2
                actions+=/howling_blast,if=dot.frost_fever.remains<=2
                actions+=/plague_strike,if=dot.blood_plague.remains<=2
                actions+=/obliterate,if=frost=2&unholy=2
                actions+=/obliterate,if=death=2
                actions+=/obliterate,if=buff.killing_machine.react
                actions+=/empower_rune_weapon,if=target.time_to_die<=120&buff.killing_machine.react
                actions+=/frost_strike,if=runic_power>=90&!buff.bloodlust.react
                actions+=/frost_strike,if=runic_power>=95
                actions+=/howling_blast,if=buff.rime.react
                actions+=/howling_blast,if=(death+unholy)=0&!buff.bloodlust.react
                actions+=/obliterate
                actions+=/empower_rune_weapon,if=target.time_to_die<=45
                actions+=/frost_strike
                actions+=/howling_blast
                actions+=/blood_tap
                actions+=/empower_rune_weapon
                actions+=/horn_of_winter
                 */


                AutoAttack();

                // Global cooldown restricted abilitiy's here
                if (SpellManager.GlobalCooldownLeft.Seconds < 0.5)
                {
                    // keep blood tap rolling.
                    if (!Me.Auras.ContainsKey("Blood Tap") && SpellManager.CanCast("Blood Tap") && Me.DeathRuneCount != 2)
                    {
                        if (logDebug)
                        {
                            logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Blood Tap", 0, 0, IsOnCooldown("Obliterate"), false);
                        }

                        SpellManager.Cast("Blood Tap");
                        return;
                    }

                    //Pillar of Frost - 1 Frost / 1 minute cooldown / Strength increased by 20%. Immune to movement from external sources. 20 secs
                    if ((Me.FrostRuneCount > 0 && Me.DeathRuneCount > 0) && (IsDebuffOnTarget("Frost Fever") && IsDebuffOnTarget("Blood Plague")) && !IsOnCooldown("Pillar of Frost"))
                    {
                        if (logDebug)
                        {
                            logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Pillar of Frost", 0, 0, false, false);
                        }

                        SpellManager.Cast("Pillar of Frost");
                        return;

                    }
                    //Raise Dead - 30 yd range / 3 min cooldown / 1 min duration
                    if (Me.HasAura("Pillar of Frost") && PlayerBuffTimeLeft("Pillar of Frost") < 3 && !IsOnCooldown("Raise Dead") && Me.CurrentTarget.HealthPercent > 20) // raise the ghoul just before pillar of frost ends to maximise dps. (Thanks imdasandman)
                    {
                        if (logDebug)
                        {
                            logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Raise Dead", PlayerBuffTimeLeft("Pillar of Frost"), 0, IsOnCooldown("Obliterate"),false);
                        }

                        SpellManager.Cast("Raise Dead");
                        return;
                    }
                    
                    //Outbreak to apply Frost Fever and Blood Plague - instant / 30 yd range 1 minute cooldown
                    if (!IsOnCooldown("Outbreak") && !IsDebuffOnTarget("Blood Plague") && IsDistanceLessThan(MaxDistance("Outbreak")) && Me.CurrentTarget.HealthPercent > 20)
                    {
                        if (logDebug)
                        {
                            logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Outbreak", 0, 0, IsOnCooldown("Obliterate"),false);
                        }

                        SpellManager.Cast("Outbreak");
                        return;

                    }

                    //Obliterate  - 1 Unholy & 1 Frost / melee range
                    if (((Me.FrostRuneCount == 2 && Me.UnholyRuneCount == 2) || Me.DeathRuneCount == 2) && IsWithinInteractRange)
                    {
                        if (logDebug)
                        {
                            logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Obliterate", 0, 0, IsOnCooldown("Obliterate"),false);
                        }


                        if (BuffExists("Killing Machine"))
                        {
                            SpellManager.Cast("Obliterate");
                            return;
                        }
                        else
                        {
                            SpellManager.Cast("Obliterate");
                            return;
                        }
                        
                    }
                    else
                    {
                        //Frost Strike  - 40 Runic Power / melee range
                        if (Me.RunicPowerPercent >= 95 && IsWithinInteractRange)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Frost Strike", 0, 0, IsOnCooldown("Obliterate"), false);
                            }

                            SpellManager.Cast("Frost Strike");
                            return;
                        }
                        //Plague Strike to apply Blood Plague - 1 Unholy / melee range
                        if ((Me.UnholyRuneCount > 0 || Me.DeathRuneCount > 0) && !IsOnCooldown("Plague Strike") && IsWithinInteractRange && Me.CurrentTarget.HealthPercent > 20)
                        {
                            if (TargetDebuffTimeLeft("Blood Plague") <= 2 || !IsDebuffOnTarget("Blood Plague"))
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Plague Strike", 0, TargetDebuffTimeLeft("Blood Plague"), IsOnCooldown("Obliterate"), false);
                                }

                                SpellManager.Cast("Plague Strike");
                                return;
                            }
                        }

                        //Howling Blast to apply Frost Fever - 1 Frost / 20 yd range
                        if ((Me.FrostRuneCount > 0 || Me.DeathRuneCount > 0) && !IsOnCooldown("Howling Blast") && IsDistanceLessThan(20) && Me.CurrentTarget.HealthPercent > 20)
                        {
                            if (TargetDebuffTimeLeft("Frost Fever") <= 2 || !IsDebuffOnTarget("Frost Fever"))
                            {

                                int countOfControlledMobs = 0;

                                foreach (WoWUnit unit in ObjectManager.GetObjectsOfType<WoWUnit>())
                                {
                                    if (unit.Distance < 20) continue;
                                    if (!unit.IsPet) continue;
                                    if (!unit.IsNonCombatPet) continue;
                                    if (!unit.IsCritter) continue;
                                    if (Me.CurrentTargetGuid == unit.Guid) continue;
                                    if (IsCrowdControlled(unit)) continue;
                                    countOfControlledMobs = +1;
                                }

                                //Logging.Write(Color.Red, "[+] Controlled Mobs = " + countOfControlledMobs);

                                if (countOfControlledMobs == 0)
                                {
                                    if (logDebug)
                                    {
                                        logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Howling Blast", 0, 0, IsOnCooldown("Obliterate"),false);
                                    }

                                    SpellManager.Cast("Howling Blast");
                                    return;

                                }
                                else
                                {
                                    if (logDebug)
                                    {
                                        logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Icy Touch", 0, 0, IsOnCooldown("Obliterate"),false);
                                    }
                                    
                                    SpellManager.Cast("Icy Touch");
                                    return;
                                }
                            }
                        }

                        //Obliterate  - 1 Unholy & 1 Frost / melee range
                        if (((Me.FrostRuneCount > 0 && Me.UnholyRuneCount > 0) || Me.DeathRuneCount >= 2) && IsWithinInteractRange && !IsOnCooldown("Obliterate"))
                        {
                            if (logDebug)
                            {
                                logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Obliterate", 0, 0, IsOnCooldown("Obliterate"),false);
                            }

                            if (BuffExists("Killing Machine"))
                            {                                                                
                                SpellManager.Cast("Obliterate");
                                return;
                            }
                            else
                            {                               
                                SpellManager.Cast("Obliterate");
                                return;
                            }
                        }

                        //Frost Strike  - 40 Runic Power / melee range
                        if (Me.RunicPowerPercent > 40 && IsWithinInteractRange)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Frost Strike", 0, 0, IsOnCooldown("Obliterate"),false);
                            }

                            SpellManager.Cast("Frost Strike");
                            return;
                        }

                        //Howling Blast + Freezing Fog - 1 Frost / 20 yards
                        if (IsDistanceLessThan(20) && BuffExists("Freezing Fog"))
                        {
                            int countOfControlledMobs = 0;

                            foreach (WoWUnit unit in ObjectManager.GetObjectsOfType<WoWUnit>())
                            {
                                if (unit.Distance < 20) continue;
                                if (!unit.IsPet) continue;
                                if (!unit.IsNonCombatPet) continue;
                                if (!unit.IsCritter) continue;
                                if (Me.CurrentTargetGuid == unit.Guid) continue;
                                if (IsCrowdControlled(unit)) continue;
                                countOfControlledMobs = +1;
                            }

                            //Logging.Write(Color.Red, "[+] Controlled Mobs = " + countOfControlledMobs);

                            if (countOfControlledMobs == 0)
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Howling Blast", 0, 0, IsOnCooldown("Obliterate"), false);
                                }

                                SpellManager.Cast("Howling Blast");
                                return;

                            }
                        }
                       
                    }

                } // end global spell cooldown
            }
        }

        public void AoERotation()
        {
            // AoE target PvE dps rotation

            if (Presence == 1) //Blood Presence AoE target
            {
                
                if (SpellManager.GlobalCooldownLeft.Seconds < 0.5)
                {

                    //Death Strike  - 1 Unholy & 1 Frost / melee range
                    if (((Me.FrostRuneCount > 0 && Me.UnholyRuneCount > 0) || Me.DeathRuneCount >= 2) && IsWithinInteractRange)
                    {
                        if (logDebug)
                        {
                            logWrite(Color.Maroon, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Death Strike", 0, 0, false, BuffExists("Will of the Necropolis"));
                        }


                        if (Me.HealthPercent < 80)
                        {
                            SpellManager.Cast("Death Strike");
                            return;
                        }
                        else if (Me.HealthPercent > 80 && BuffExists("Blood Shield") && PlayerBuffTimeLeft("Blood Shield") < 5)
                        {
                            SpellManager.Cast("Death Strike");
                            return;
                        }
                        else
                        {
                            SpellManager.Cast("Death Strike");
                            return;
                        }
                    }
                    else
                    {

                        //Death and Decay - 1 Unholy / melee range -- (!IsDebuffOnTarget("Blood Plague") || !IsDebuffOnTarget("Scarlet Fever"))
                        if ((Me.UnholyRuneCount > 0 || Me.DeathRuneCount > 0) && !IsOnCooldown("Death and Decay") && Me.HealthPercent > 30)
                        {

                            int countOfControlledMobs = 0;

                            foreach (WoWUnit unit in ObjectManager.GetObjectsOfType<WoWUnit>())
                            {
                                if (unit.Distance < 10) continue;
                                if (!unit.IsPet) continue;
                                if (!unit.IsNonCombatPet) continue;
                                if (!unit.IsCritter) continue;
                                if (Me.CurrentTargetGuid == unit.Guid) continue;
                                if (IsCrowdControlled(unit)) continue;
                                countOfControlledMobs = +1;
                            }

                            //Logging.Write(Color.Lime, "[+] Controlled Mobs = " + countOfControlledMobs);

                            if (countOfControlledMobs == 0)
                            {
                                if (Cast("Death and Decay", Me.CurrentTarget.Location))
                                {
                                    if (logDebug)
                                    {
                                        logWrite(Color.Maroon, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Death and Decay", 0, 0, false, BuffExists("Will of the Necropolis"));
                                    }
                                    return;
                                }
                            }

                        }


                        //Outbreak to apply Frost Fever and Blood Plague and Scarlet Fever - instant / 30 yd range 1 minute cooldown
                        if (!IsOnCooldown("Outbreak") && (!IsDebuffOnTarget("Frost Fever") && !IsDebuffOnTarget("Blood Plague")) && IsDistanceLessThan(MaxDistance("Outbreak")) && Me.CurrentTarget.HealthPercent > 20)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Maroon, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Outbreak", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }
                            
                            SpellManager.Cast("Outbreak");
                            return;

                        }

                        //Icy Touch to apply Frost Fever - 1 Frost / 20 yd range
                        if ((Me.FrostRuneCount > 0 || Me.DeathRuneCount > 0) && !IsDebuffOnTarget("Frost Fever") && IsDistanceLessThan(20) && Me.CurrentTarget.HealthPercent > 20 && Me.HealthPercent > 30)
                        {

                            if (TargetDebuffTimeLeft("Frost Fever") <= 4 || !IsOnCooldown("Icy Touch"))
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.Maroon, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Icy Touch", 0, 0, false, BuffExists("Will of the Necropolis"));
                                }

                                SpellManager.Cast("Icy Touch");
                                return;
                            }

                        }

                        //Plague Strike to apply Blood Plague and Scarlet Fever - 1 Unholy / melee range -- (!IsDebuffOnTarget("Blood Plague") || !IsDebuffOnTarget("Scarlet Fever"))
                        if ((Me.UnholyRuneCount > 0 || Me.DeathRuneCount > 0) && !IsDebuffOnTarget("Blood Plague") && IsWithinInteractRange && Me.CurrentTarget.HealthPercent > 20 && Me.HealthPercent > 30)
                        {
                            if (TargetDebuffTimeLeft("Blood Plague") <= 4 || !IsOnCooldown("Plague Strike"))
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.Maroon, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Plague Strike", 0, 0, false, BuffExists("Will of the Necropolis"));
                                }

                                SpellManager.Cast("Plague Strike");
                                return;
                            }
                        }

                        // [Pestilence] to spread Frost Fever and Blood Plague and Scarlet Fever - instant / cost 1 Blood / 30 yd range 1 minute cooldown
                        if ((Me.BloodRuneCount > 1 || Me.DeathRuneCount > 0) && !IsOnCooldown("Pestilence") && Me.HealthPercent > 30)
                        {
                            int countOfDiseaselessMobs = 0;

                            foreach (WoWUnit unit in ObjectManager.GetObjectsOfType<WoWUnit>())
                            {
                                if (unit.Distance < 10) continue;
                                if (!unit.IsPet) continue;
                                if (!unit.IsNonCombatPet) continue;
                                if (!unit.IsCritter) continue;
                                if (Me.CurrentTargetGuid == unit.Guid) continue;
                                if (unit.Auras.ContainsKey("Blood Plague")) continue;
                                if (unit.Auras.ContainsKey("Frost Fever")) continue;
                                countOfDiseaselessMobs = +1;
                            }

                            //Logging.Write(Color.Lime, "[+] Diseaseless Mobs = " + countOfDiseaselessMobs);

                            if (countOfDiseaselessMobs == 0)
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.Maroon, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Pestilence", 0, 0, false, BuffExists("Will of the Necropolis"));
                                }
                                SpellManager.Cast("Pestilence");
                                return;
                            }

                        }

                        //Heart Strike  - 1 Blood / melee range
                        if ((Me.BloodRuneCount > 1 || Me.DeathRuneCount > 0) && !IsOnCooldown("Heart Strike") && IsWithinInteractRange)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Maroon, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Heart Strike", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }
                            
                            SpellManager.Cast("Heart Strike");
                            return;
                        }
                        if ((Me.BloodRuneCount > 1 || Me.DeathRuneCount > 0) && IsDistanceLessThan(8))
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Maroon, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Blood Boil", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }

                            SpellManager.Cast("Blood Boil");
                            return;
                        }

                        //Rune Strike  - 30 Runic Power / melee range
                        if (Me.RunicPowerPercent > 25 && IsWithinInteractRange && Me.HealthPercent > 30)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Maroon, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Rune Strike", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }

                            SpellManager.Cast("Rune Strike");
                            return;
                        }

                        //Death Coil  - 40 Runic Power / 30 yards
                        else if (Me.RunicPowerPercent > 35 && IsDistanceLessThan(MaxDistance("Death Coil")) && !BuffExists("Lichborne") && Me.HealthPercent > 30)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.Maroon, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Death Coil", 0, 0, false, BuffExists("Will of the Necropolis"));
                            }

                            SpellManager.Cast("Death Coil");
                            return;
                        }
                    }
                }
            }
            if (Presence == 0)//Unholy Presence Frost spec uses Unholy AoE target
            {
                /*
                  glyphs=obliterate/frost_strike/howling_blast/pestilence/blood_boil/dark_succor/horn_of_winter/blood_tap/path_of_frost
                  actions=flask,type=titanic_strength
                  actions+=/food,type=beer_basted_crocolisk
                  actions+=/presence,choose=unholy
                  actions+=/army_of_the_dead
                  actions+=/snapshot_stats
                  actions+=/golemblood_potion,if=!in_combat|buff.bloodlust.react|target.time_to_die<=60
                  actions+=/auto_attack
                  actions+=/pillar_of_frost
                  actions+=/blood_tap,if=death!=2
                  actions+=/raise_dead,if=buff.rune_of_the_fallen_crusader.react&buff.heart_of_rage.react
                  actions+=/raise_dead,time>=15
                  actions+=/outbreak,if=dot.frost_fever.remains<=2|dot.blood_plague.remains<=2
                  actions+=/howling_blast,if=dot.frost_fever.remains<=2
                  actions+=/plague_strike,if=dot.blood_plague.remains<=2
                  actions+=/obliterate,if=frost=2&unholy=2
                  actions+=/obliterate,if=death=2
                  actions+=/obliterate,if=buff.killing_machine.react
                  actions+=/empower_rune_weapon,if=target.time_to_die<=120&buff.killing_machine.react
                  actions+=/frost_strike,if=runic_power>=90&!buff.bloodlust.react
                  actions+=/frost_strike,if=runic_power>=95
                  actions+=/howling_blast,if=buff.rime.react
                  actions+=/howling_blast,if=(death+unholy)=0&!buff.bloodlust.react
                  actions+=/obliterate
                  actions+=/empower_rune_weapon,if=target.time_to_die<=45
                  actions+=/frost_strike
                  actions+=/howling_blast
                  actions+=/blood_tap
                  actions+=/empower_rune_weapon
                  actions+=/horn_of_winter
                   */


                AutoAttack();

                // Global cooldown restricted abilitiy's here
                if (SpellManager.GlobalCooldownLeft.Seconds < 0.5)
                {
                    // keep blood tap rolling.
                    if (!Me.Auras.ContainsKey("Blood Tap") && SpellManager.CanCast("Blood Tap") && Me.DeathRuneCount != 2)
                    {
                        if (logDebug)
                        {
                            logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Blood Tap", 0, 0, IsOnCooldown("Obliterate"), false);
                        }

                        SpellManager.Cast("Blood Tap");
                        return;
                    }

                    //Pillar of Frost - 1 Frost / 1 minute cooldown / Strength increased by 20%. Immune to movement from external sources. 20 secs
                    if ((Me.FrostRuneCount > 0 && Me.DeathRuneCount > 0) && (IsDebuffOnTarget("Frost Fever") && IsDebuffOnTarget("Blood Plague")) && !IsOnCooldown("Pillar of Frost"))
                    {
                        if ((SpellManager.CanCast("Pillar of Frost")))
                        {
                            if (logDebug)
                            {
                                logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Pillar of Frost", 0, 0, IsOnCooldown("Obliterate"),false);
                            }

                            SpellManager.Cast("Pillar of Frost");
                            return;

                        }
                    }

                    //Outbreak to apply Frost Fever and Blood Plague - instant / 30 yd range 1 minute cooldown
                    if (!IsOnCooldown("Outbreak") && !IsDebuffOnTarget("Blood Plague") && IsDistanceLessThan(MaxDistance("Outbreak")) && Me.CurrentTarget.HealthPercent > 20)
                    {
                        if (logDebug)
                        {
                            logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Outbreak", 0, 0, IsOnCooldown("Obliterate"),false);
                        }

                        SpellManager.Cast("Outbreak");
                        return;
                    }

                    //Obliterate  - 1 Unholy & 1 Frost / melee range
                    if (((Me.FrostRuneCount == 2 && Me.UnholyRuneCount == 2) || Me.DeathRuneCount == 2) && IsWithinInteractRange)
                    {
                        if (logDebug)
                        {
                            logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Obliterate", 0, 0, IsOnCooldown("Obliterate"),false);
                        }

                        if (BuffExists("Killing Machine"))
                        {
                            SpellManager.Cast("Obliterate");
                            return;
                        }
                        else
                        {
                            SpellManager.Cast("Obliterate");
                            return;
                        }

                    }
                    else
                    {
                        //Frost Strike  - 40 Runic Power / melee range
                        if (Me.RunicPowerPercent >= 95 && IsWithinInteractRange)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.LightBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Frost Strike", 0, 0, IsOnCooldown("Obliterate"), false);
                            }

                            SpellManager.Cast("Frost Strike");
                            return;
                        }
                        //Howling Blast + Freezing Fog - 1 Frost / 20 yards
                        if (IsDistanceLessThan(20) && BuffExists("Freezing Fog"))
                        {
                            int countOfControlledMobs = 0;

                            foreach (WoWUnit unit in ObjectManager.GetObjectsOfType<WoWUnit>())
                            {
                                if (unit.Distance < 20) continue;
                                if (!unit.IsPet) continue;
                                if (!unit.IsNonCombatPet) continue;
                                if (!unit.IsCritter) continue;
                                if (Me.CurrentTargetGuid == unit.Guid) continue;
                                if (IsCrowdControlled(unit)) continue;
                                countOfControlledMobs = +1;
                            }

                            //Logging.Write(Color.Red, "[+] Controlled Mobs = " + countOfControlledMobs);

                            if (countOfControlledMobs == 0)
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Howling Blast", 0, 0, IsOnCooldown("Obliterate"),false);
                                }

                                SpellManager.Cast("Howling Blast");
                                return;

                            }
                        }
                        //Plague Strike to apply Blood Plague - 1 Unholy / melee range
                        if ((Me.UnholyRuneCount > 0 || Me.DeathRuneCount > 0) && !IsOnCooldown("Plague Strike") && IsWithinInteractRange && Me.CurrentTarget.HealthPercent > 20)
                        {
                            if (TargetDebuffTimeLeft("Blood Plague") <= 3 || !IsDebuffOnTarget("Blood Plague"))
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Plague Strike", 0, TargetDebuffTimeLeft("Blood Plague"), IsOnCooldown("Obliterate"), false);
                                }

                                SpellManager.Cast("Plague Strike");
                                return;
                            }
                        }

                        //Howling Blast to apply Frost Fever - 1 Frost / 20 yd range
                        if ((Me.FrostRuneCount > 0 || Me.DeathRuneCount > 0) && !IsOnCooldown("Howling Blast") && IsDistanceLessThan(20) && Me.CurrentTarget.HealthPercent > 20)
                        {
                            if (TargetDebuffTimeLeft("Frost Fever") <= 4 || !IsDebuffOnTarget("Frost Fever"))
                            {

                                int countOfControlledMobs = 0;

                                foreach (WoWUnit unit in ObjectManager.GetObjectsOfType<WoWUnit>())
                                {
                                    if (unit.Distance < 20) continue;
                                    if (!unit.IsPet) continue;
                                    if (!unit.IsNonCombatPet) continue;
                                    if (!unit.IsCritter) continue;
                                    if (Me.CurrentTargetGuid == unit.Guid) continue;
                                    if (IsCrowdControlled(unit)) continue;
                                    countOfControlledMobs = +1;
                                }

                                //Logging.Write(Color.Red, "[+] Controlled Mobs = " + countOfControlledMobs);

                                if (countOfControlledMobs == 0)
                                {
                                    if (logDebug)
                                    {
                                        logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Howling Blast", 0, 0, IsOnCooldown("Obliterate"),false);
                                    }

                                    SpellManager.Cast("Howling Blast");
                                    return;

                                }
                                else
                                {
                                    if (logDebug)
                                    {
                                        logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Icy Touch", 0, 0, IsOnCooldown("Obliterate"),false);
                                    }

                                    SpellManager.Cast("Icy Touch");
                                    return;
                                }
                            }
                        }
                        if ((Me.FrostRuneCount > 0 || Me.DeathRuneCount > 0) && !IsOnCooldown("Howling Blast") && IsDistanceLessThan(20) && Me.CurrentTarget.HealthPercent > 20)
                        {
                            int countOfControlledMobs = 0;

                            foreach (WoWUnit unit in ObjectManager.GetObjectsOfType<WoWUnit>())
                            {
                                if (unit.Distance < 20) continue;
                                if (!unit.IsPet) continue;
                                if (!unit.IsNonCombatPet) continue;
                                if (!unit.IsCritter) continue;
                                if (Me.CurrentTargetGuid == unit.Guid) continue;
                                if (IsCrowdControlled(unit)) continue;
                                countOfControlledMobs = +1;
                            }

                            //Logging.Write(Color.Red, "[+] Controlled Mobs = " + countOfControlledMobs);

                            if (countOfControlledMobs == 0)
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Howling Blast", 0, 0, IsOnCooldown("Obliterate"), false);
                                }

                                SpellManager.Cast("Howling Blast");
                                return;

                            }
                        }

                        //Obliterate  - 1 Unholy & 1 Frost / melee range
                        if (((Me.FrostRuneCount > 0 && Me.UnholyRuneCount > 0) || Me.DeathRuneCount >= 2) && IsWithinInteractRange && !IsOnCooldown("Obliterate"))
                        {
                            if (logDebug)
                            {
                                logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Obliterate", 0, 0, IsOnCooldown("Obliterate"),false);
                            }

                            if (BuffExists("Killing Machine"))
                            {
                                SpellManager.Cast("Obliterate");
                                return;
                            }
                            else
                            {                              
                                SpellManager.Cast("Obliterate");
                                return;
                            }
                        }

                        //Frost Strike  - 40 Runic Power / melee range
                        if (Me.RunicPowerPercent > 40 && IsWithinInteractRange)
                        {
                            if (logDebug)
                            {
                                logWrite(Color.MediumBlue, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Frost Strike", 0, 0, IsOnCooldown("Obliterate"),false);
                            }

                            SpellManager.Cast("Frost Strike");
                            return;
                        }

                    }
                } // end global spell cooldown
            }            
        }
       
        public void Cooldowns()
        {
            // handles cooldown usage, and stacking cooldowns

            if (UseCooldowns != 0)
            {
                // 
                if (!SpellManager.GlobalCooldown)
                {
                    if (UseCooldowns == 1 || (!Me.IsInRaid && IsTargetDungeonBoss(Me.CurrentTarget))
                        || IsTargetRaidBoss())
                    {

                        /////////////////////////////////////////////////////////////////////////////////////////////////////////////
                        /////     Blood Tank Specific
                        /////////////////////////////////////////////////////////////////////////////////////////////////////////////
                        if (Presence == 1) //Blood Presence AoE target
                        {
                            // keep blood tap rolling.
                            if (!Me.Auras.ContainsKey("Blood Tap") && Me.Auras.ContainsKey("Bone Shield") && SpellManager.CanCast("Blood Tap"))
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Blood Tap", 0, 0, false, BuffExists("Will of the Necropolis"));
                                }                                

                                SpellManager.Cast("Blood Tap");                                
                                return;
                            }

                            // Empower Rune Weapon / 5 min cooldown / Instant / Empower your rune weapon, immediately activating all your runes and generating 25 Runic Power.
                            else if ((Me.BloodRuneCount + Me.FrostRuneCount + Me.UnholyRuneCount + Me.DeathRuneCount < 3) && SpellManager.CanCast("Empower Rune Weapon") && UseERW)
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Empower Rune Weapon", 0, 0, false, BuffExists("Will of the Necropolis"));
                                }

                                SpellManager.Cast("Empower Rune Weapon");
                                return;
                            }

                            // Dancing Rune Weapon / 1.5 minutes cooldown / Instant / 30 yards / Summons a second rune weapon that fights on its own for 12 sec, mirroring the Death Knight's attacks.  The rune weapon also assists in defense of its master, granting an additional 20% parry chance while active.
                            else if (Me.CurrentRunicPower > 60 && SpellManager.CanCast("Dancing Rune Weapon") && IsKnown("Dancing Rune Weapon") && Me.HealthPercent > 30 && UseDRW)
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, false, false, BuffExists("Blood Shield"), Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Dancing Rune Weapon", 0, 0, false, BuffExists("Will of the Necropolis"));
                                }

                                SpellManager.Cast("Dancing Rune Weapon");
                                return;
                            }

                        }
                        /////////////////////////////////////////////////////////////////////////////////////////////////////////////
                        /////     Frost Dps specific
                        /////////////////////////////////////////////////////////////////////////////////////////////////////////////
                        if (Presence == 0)//Unholy Presence Frost spec uses Unholy AoE target
                        {
                            // Empower Rune Weapon / 5 min cooldown / Instant / Empower your rune weapon, immediately activating all your runes and generating 25 Runic Power.
                            if ((Me.BloodRuneCount + Me.FrostRuneCount + Me.UnholyRuneCount + Me.DeathRuneCount < 3) && SpellManager.CanCast("Empower Rune Weapon") && UseERW)
                            {
                                if (logDebug)
                                {
                                    logWrite(Color.Lime, Me.CurrentTarget.Name, Me.HealthPercent, BuffExists("Freezing Fog"), BuffExists("Killing Machine"), false, Me.FrostRuneCount, Me.UnholyRuneCount, Me.BloodRuneCount, Me.DeathRuneCount, Me.CurrentRunicPower, Me.CurrentTarget.Distance, "Empower Rune Weapon", 0, 0, IsOnCooldown("Obliterate"),false);
                                }

                                SpellManager.Cast("Empower Rune Weapon");
                                return;
                            }
                        }    
                    }

                }
            }
        }

        public void TrinketUsage()
        {
            // shout-out to wulf for the code he sent me
            // regarding trinket usage

            var Trinket1 = StyxWoW.Me.Inventory.GetItemBySlot(13);
            var Trinket2 = StyxWoW.Me.Inventory.GetItemBySlot(12);

            /*
            var slot1 = StyxWoW.Me.Inventory.GetItemBySlot(1); //Neck
            var slot2 = StyxWoW.Me.Inventory.GetItemBySlot(2); //Shoulders
            var slot3 = StyxWoW.Me.Inventory.GetItemBySlot(3); //shirt
            var slot4 = StyxWoW.Me.Inventory.GetItemBySlot(4); //chest
            var slot5 = StyxWoW.Me.Inventory.GetItemBySlot(5); //belt
            var slot6 = StyxWoW.Me.Inventory.GetItemBySlot(6); //legs
            var slot7 = StyxWoW.Me.Inventory.GetItemBySlot(7); //boots
            var slot8 = StyxWoW.Me.Inventory.GetItemBySlot(8); //wrist
            var slot9 = StyxWoW.Me.Inventory.GetItemBySlot(9); //hands
            var slot10 = StyxWoW.Me.Inventory.GetItemBySlot(10); //ring 1
            var slot11 = StyxWoW.Me.Inventory.GetItemBySlot(11); //ring 2
            var slot12 = StyxWoW.Me.Inventory.GetItemBySlot(12); //Bottom Trinket (2)
            var slot12 = StyxWoW.Me.Inventory.GetItemBySlot(13); //Top Trinket (1)
            var slot12 = StyxWoW.Me.Inventory.GetItemBySlot(14); //cloak

            Logging.Write(Color.Yellow, "slot1 = " + slot1.Name);
            Logging.Write(Color.Yellow, "slot2 = " + slot2.Name);
            Logging.Write(Color.Yellow, "slot3 = " + slot3.Name);
            Logging.Write(Color.Yellow, "slot4 = " + slot4.Name);
            Logging.Write(Color.Yellow, "slot5 = " + slot5.Name);
            Logging.Write(Color.Yellow, "slot6 = " + slot6.Name);
            Logging.Write(Color.Yellow, "slot7 = " + slot7.Name);
            Logging.Write(Color.Yellow, "slot8 = " + slot8.Name);
            Logging.Write(Color.Yellow, "slot9 = " + slot9.Name);
            Logging.Write(Color.Yellow, "slot10 = " + slot10.Name);
            Logging.Write(Color.Yellow, "slot11 = " + slot11.Name);
            Logging.Write(Color.Yellow, "slot12 = " + slot12.Name);
            Logging.Write(Color.Yellow, "slot12 = " + slot13.Name);
            Logging.Write(Color.Yellow, "slot12 = " + slot14.Name);
            
             */


            if (Trinket1 != null && Trinket1.Usable && Trinket1.Cooldown <= 0)
            {
                if (UseTrinket1 == 1)
                {
                    Logging.Write(Color.Yellow, "=== Using trinket: " + Trinket1.Name);
                    Trinket1.Use();
                }
                else if (UseTrinket1 == 2)
                {
                    Logging.Write(Color.Yellow, "=== Using trinket: " + Trinket1.Name);
                    Trinket1.Use();
                }
            }


            if (Trinket2 != null && Trinket2.Usable && Trinket2.Cooldown <= 0)
            {
                if (UseTrinket2 == 1)
                {
                    Logging.Write(Color.Yellow, "=== Using trinket: " + Trinket2.Name);
                    Trinket2.Use();
                }
                else if (UseTrinket2 == 2)
                {
                    Logging.Write(Color.Yellow, "=== Using trinket: " + Trinket2.Name);
                    Trinket2.Use();
                }
            }
        }

        public double SpellCooldown(string SpellName)
        {
            try
            {
                // returns spell cooldown in seconds
                string CooldownLua = string.Format("local start, duration = GetSpellCooldown(\"{0}\"); if (start+duration)==nil then return 0 else return (start+duration) - GetTime() end", Lua.Escape(SpellName));
                //Logging.Write(Color.Red, SpellName + " cooldown: " + Lua.GetReturnValues(CooldownLua)[0]);
                return double.Parse(Lua.GetReturnValues(CooldownLua)[0]);
            }
            catch
            {
                return 999999;
            }
        }

        public double TargetDebuffTimeLeft(string name)
        {
            try
            {
                var lua = string.Format("local x=select(7, UnitDebuff(\"target\", \"{0}\", nil, 'PLAYER')); if x==nil then return 0 else return x-GetTime() end", Lua.Escape(name));
                var t = double.Parse(Lua.GetReturnValues(lua)[0]);
                return t;
            }
            catch
            {
                return 999999;
            }
        }
        
        public double PlayerBuffTimeLeft(string name)
        {
            try
            {
                var lua = string.Format("local x=select(7, UnitBuff('player', \"{0}\", nil, 'PLAYER')); if x==nil then return 0 else return x-GetTime() end", Lua.Escape(name));
                var t = double.Parse(Lua.GetReturnValues(lua)[0]);
                return t;
            }
            catch
            {
                return 999999;
            }
        }

        public bool BuffExists(string BuffName)
        {
            // returns true if buff is on the player, else false
            string BuffExistsLua = string.Format("local name = UnitBuff(\"player\", \"{0}\"); " +
                                                 "return name", BuffName);
            string BuffNameLua = Lua.GetReturnValues(BuffExistsLua)[0];

            if (BuffNameLua == BuffName) return true;
            else return false;
        }

        public bool InterruptibleCasting(string unit)
        {

            int CastInterruptibleLua = Lua.GetReturnVal<int>("local name, subText, text, texture, startTime, endTime, isTradeSkill, DcastID, notInterruptible = UnitCastingInfo(" + unit + "); return notInterruptible", 0);
            //Logging.Write(Color.Red, unit.Guid + " can interupt = " + CastInterruptibleLua);
            if (CastInterruptibleLua == 0) return true;
            else return false;
            
        }

        public bool InterruptibleChannel(string unit)
        {
            bool ChanInterruptibleLua = Lua.GetReturnVal<bool>("local name, subText, text, texture, startTime, endTime, isTradeSkill, DcastID, notInterruptible = UnitCastingInfo(" + unit + "); return notInterruptible", 0);
            //Logging.Write(Color.Red, unit.Guid + " can interupt = " + ChanInterruptibleLua);
            if (!ChanInterruptibleLua) return true;
            else return false;
        }

        public bool IsDistanceLessThan(double distanceCheck) { return Me.GotTarget && Me.CurrentTarget.Distance < distanceCheck; }

        public bool IsDistanceMoreThan(double distanceCheck) { return Me.GotTarget && Me.CurrentTarget.Distance > distanceCheck; }

        public static bool IsWithinInteractRange { get { return Me.GotTarget && Me.CurrentTarget.WithinInteractRange; } }

        public double MaxDistance(string spellName)
        {
            if (!IsKnown(spellName)) return 0.0;

            return SpellManager.Spells[spellName].MaxRange;
        }

        public bool IsKnown(string spellName) { return SpellManager.HasSpell(spellName); }

        public int PowerCost(string spellName)
        {
            if (!IsKnown(spellName)) return 9999999;

            return SpellManager.Spells[spellName].PowerCost;
        }
        
        public bool IsOnCooldown(string spellName)
        {
            if (!IsKnown(spellName)) return true;

            bool result = SpellManager.Spells[spellName].Cooldown;

            return result;
        }

        public bool IsDebuffOnTarget(string debuffName) { return Me.GotTarget && Me.CurrentTarget.HasAura(debuffName) && Me.CurrentTarget.Auras[debuffName].CreatorGuid == Me.Guid; } 
        
        public static WoWUnit PartyMemberWithoutBuff(string buffName)
        {
            // If you're not in a party then just leave
            if (!Me.IsInParty && !Me.IsInRaid) return null;
            return Me.PartyMembers.FirstOrDefault(p => p.Distance < 40 && !p.Dead && !p.IsGhost && p.InLineOfSight && !p.Auras.ContainsKey(buffName));
        }

        public void LoadGlyphs()
        {
            //glyph handling have you got that glyph ??
            var glyphCount = Lua.GetReturnVal<int>("return GetNumGlyphSockets()", 0);
            
            Glyphs = new HashSet<string>();

            if (glyphCount != 0)
            {
                for (int i = 1; i <= glyphCount; i++)
                {
                    List<string> glyphInfo = Lua.GetReturnValues(String.Format("return GetGlyphSocketInfo({0})", i));

                    if (glyphInfo != null && glyphInfo[3] != "nil" && !string.IsNullOrEmpty(glyphInfo[3]))
                    {
                        Glyphs.Add(WoWSpell.FromId(int.Parse(glyphInfo[3])).Name.Replace("Glyph of ", ""));
                    }
                }
            }
        }

        public bool HasGlyph(string glyphName)
        {
            return Glyphs.Count > 0 && Glyphs.Contains(glyphName);
        }

        public bool IsTargetRaidBoss()
        {
            // returns true is target is boss, false if not
            string UnitClassificationLua = string.Format("local classification = UnitClassification(\"target\"); return classification");
            string UnitClassification = Lua.GetReturnValues(UnitClassificationLua)[0];

            if (UnitClassification == "worldboss") return true;
            else return false;
        }

        public bool IsTargetDungeonBoss(WoWUnit target)
        {
            // checks if target is a dungeon boss
            if (target.Level == 87 && target.Elite) return true;
            else return false;
        }

        public bool CastSpell(int spellID)
        {
            if (SpellManager.CanCast(spellID) && !SpellManager.GlobalCooldown && !Me.IsCasting)
            {
                SpellManager.Cast(spellID);
                // We managed to cast the spell, so return true, saying we were able to cast it.
                return true;
            }
            // Can't cast the spell right now, so return false.
            return false;
        }

        public bool Cast(string spellName, WoWPoint clickCastLocation)
        {

            bool result = SpellManager.Cast(spellName);
            LegacySpellManager.ClickRemoteLocation(clickCastLocation);
            return result;

        }

        public void logWrite(System.Drawing.Color clr, string currentTarget, double health, bool freezingfog, bool killingmachine, bool bloodshield, int frostrunecount, int unholyrunecount, int bloodruncount, int deathruncount, uint runicpower, double targetdistance, string spellcast, double playerdebufftime, double targetdebufftime, bool isoncooldown, bool willofnecropolis) 
        {
            if (Presence == 1) //Blood Presence
            {
                Logging.Write("==============[" + currentTarget + "]==============");
                Logging.Write(clr, "=== Health: " + health);
                Logging.Write(clr, "=== Blood Shield: " + bloodshield);
                Logging.Write(clr, "=== Frost:[" + frostrunecount + "] Unholy:[" + unholyrunecount + "] Blood:[" + bloodruncount + "] Death:[" + deathruncount + "]");
                Logging.Write(clr, "=== Runic Power: " + runicpower);
                Logging.Write(clr, "=== Distance to target: " + targetdistance);
                Logging.Write(clr, "=== Spell cast: " + spellcast);
                Logging.Write(clr, "=== Will of the Necropolis: " + willofnecropolis);
                Logging.Write(clr, "=== Plague Strike had " + targetdebufftime + " Time left");
                Logging.Write("===================================================");
            }
            if (Presence == 0) //Unholy Presence - Frost spec uses Unholy presence
            {
                Logging.Write("==============[" + currentTarget + "]==============");
                Logging.Write(clr, "=== Health: " + health);
                Logging.Write(clr, "=== Freezing Fog: " + freezingfog);
                Logging.Write(clr, "=== Killing Machine: " + killingmachine);
                Logging.Write(clr, "=== Blood Shield: " + BuffExists("Blood Shield"));
                Logging.Write(clr, "=== Frost:[" + frostrunecount + "] Unholy:[" + unholyrunecount + "] Blood:[" + bloodruncount + "] Death:[" + deathruncount + "]");
                Logging.Write(clr, "=== Distance to target: " + targetdistance);
                Logging.Write(clr, "=== Spell cast: " + spellcast);
                Logging.Write(clr, "=== Obliterate on CD: " + isoncooldown);
                Logging.Write(clr, "=== Plague Strike had " + targetdebufftime + " Time left");
                Logging.Write(clr, "=== Pillar of Frost had " + playerdebufftime + " seconds left when we poped ghoul");
                Logging.Write("===================================================");

            }
        
        }
        
        public void InitialiseSettings()
        {
            CurrentSettings = "";
            
            // initialise variables for user settings
            if (Me.IsInRaid && CurrentSettings != "Raid")
            {
                logDebug = DeathstrikeSettings.Instance.Setting_Logging;
                Presence = DeathstrikeSettings.Instance.Setting_Presence;
                ForcePresence = DeathstrikeSettings.Instance.Setting_ForcePresence;
                IceboundFortitude = DeathstrikeSettings.Instance.Setting_IceboundFortitude;
                VampiricBlood = DeathstrikeSettings.Instance.Setting_VampiricBlood;
                RuneTap = DeathstrikeSettings.Instance.Setting_RuneTap;
                DeathPact = DeathstrikeSettings.Instance.Setting_DeathPact;
                LichborneDC = DeathstrikeSettings.Instance.Setting_LichborneDC;
                UseERW = DeathstrikeSettings.Instance.Setting_UseERW;
                UseDRW = DeathstrikeSettings.Instance.Setting_UseDRW;
                UseTaunt = DeathstrikeSettings.Instance.Setting_Raid_UseTaunt;
                UseNeverDismount = DeathstrikeSettings.Instance.Setting_Raid_UseNeverDismount;
                UseInterrupt = DeathstrikeSettings.Instance.Setting_Raid_UseInterrupt;
                UseTarget = DeathstrikeSettings.Instance.Setting_Raid_UseTarget;
                UseTrinket1 = DeathstrikeSettings.Instance.Setting_Raid_UseTrinket1;
                UseTrinket2 = DeathstrikeSettings.Instance.Setting_Raid_UseTrinket2;
                UseCooldowns = DeathstrikeSettings.Instance.Setting_Raid_UseCooldowns;

                CurrentSettings = "Raid";
            }
            else if (CurrentSettings != "Party")
            {
                logDebug = DeathstrikeSettings.Instance.Setting_Logging;
                Presence = DeathstrikeSettings.Instance.Setting_Presence;
                ForcePresence = DeathstrikeSettings.Instance.Setting_ForcePresence;
                IceboundFortitude = DeathstrikeSettings.Instance.Setting_IceboundFortitude;
                VampiricBlood = DeathstrikeSettings.Instance.Setting_VampiricBlood;
                RuneTap = DeathstrikeSettings.Instance.Setting_RuneTap;
                DeathPact = DeathstrikeSettings.Instance.Setting_DeathPact;
                LichborneDC = DeathstrikeSettings.Instance.Setting_LichborneDC;
                UseERW = DeathstrikeSettings.Instance.Setting_UseERW;
                UseDRW = DeathstrikeSettings.Instance.Setting_UseDRW;
                UseMovement = DeathstrikeSettings.Instance.Setting_Party_UseMovement;
                UseFaceTarget = DeathstrikeSettings.Instance.Setting_Party_UseFaceTarget;
                UseTaunt = DeathstrikeSettings.Instance.Setting_Party_UseTaunt;
                UseNeverDismount = DeathstrikeSettings.Instance.Setting_Party_UseNeverDismount;
                UseInterrupt = DeathstrikeSettings.Instance.Setting_Party_UseInterrupt;
                UseTarget = DeathstrikeSettings.Instance.Setting_Party_UseTarget;
                UseTrinket1 = DeathstrikeSettings.Instance.Setting_Party_UseTrinket1;
                UseTrinket2 = DeathstrikeSettings.Instance.Setting_Party_UseTrinket2;
                UseCooldowns = DeathstrikeSettings.Instance.Setting_Party_UseCooldowns;

                CurrentSettings = "Party";
            }
        }
  
        public void RunLUA(string sCmd)
        {
            Lua.DoString(sCmd, "Deathstrike.lua");
        }
        
    }
}
